@@ -1,20 +1,42 @@
+#![allow(clippy::cast_lossless)]
+#![allow(clippy::cast_possible_truncation)]
+#![allow(clippy::cast_sign_loss)]
+#![allow(clippy::cognitive_complexity)]
+#![allow(clippy::doc_markdown)]
+#![allow(clippy::identity_op)]
+#![allow(clippy::if_same_then_else)]
+#![allow(clippy::inline_always)]
 #![allow(clippy::manual_range_contains)]
+#![allow(clippy::many_single_char_names)]
+#![allow(clippy::new_without_default)]
+#![allow(clippy::new_without_default)]
+#![allow(clippy::similar_names)]
+#![allow(clippy::too_many_arguments)]
+#![allow(clippy::trivially_copy_pass_by_ref)]
+#![allow(clippy::type_complexity)]
+#![allow(clippy::unreadable_literal)]
+#![allow(clippy::module_name_repetitions)]
+#![allow(clippy::missing_errors_doc)]
+#![allow(clippy::verbose_bit_mask)]
+#![allow(unexpected_cfgs)]
+#![deny(clippy::unnecessary_mut_passed)]
 
 #[allow(non_camel_case_types)]
 pub mod ffi;
 
 mod rustimpl;
-use crate::rustimpl::{lodepng_chunk_length, lodepng_get_bpp_lct, lodepng_zlib_decompress};
+mod zlib;
+
+use crate::rustimpl::chunk_length;
 
 mod error;
 pub use crate::error::*;
 mod iter;
 use crate::iter::{ChunksIterFragile, ITextKeysIter, TextKeysIter};
 
-use fallible_collections::FallibleVec;
-
-pub use rgb::Pod;
-pub use rgb::RGB;
+pub use rgb::bytemuck;
+pub use rgb::bytemuck::Pod;
+pub use rgb::Rgb as RGB;
 pub use rgb::RGBA8 as RGBA;
 
 use std::cmp;
@@ -22,23 +44,35 @@ use std::convert::TryInto;
 use std::fmt;
 use std::fs;
 use std::mem;
+use std::num::NonZeroU8;
 use std::os::raw::c_uint;
 use std::os::raw::c_void;
 use std::path::Path;
 use std::ptr;
 use std::slice;
 
+#[doc(inline)]
 pub use crate::ffi::ColorType;
+#[doc(inline)]
 pub use crate::ffi::CompressSettings;
+#[doc(inline)]
 pub use crate::ffi::DecoderSettings;
+#[doc(inline)]
 pub use crate::ffi::DecompressSettings;
+#[doc(inline)]
 pub use crate::ffi::EncoderSettings;
+#[doc(hidden)]
 pub use crate::ffi::ErrorCode;
+#[doc(inline)]
 pub use crate::ffi::FilterStrategy;
+#[doc(hidden)]
 pub use crate::ffi::State;
+#[doc(inline)]
 pub use crate::ffi::Time;
 
+#[doc(inline)]
 pub use crate::ffi::ColorMode;
+#[doc(inline)]
 pub use crate::ffi::Info;
 
 pub use iter::ChunksIter;
@@ -52,7 +86,7 @@ impl ColorMode {
 
     #[inline(always)]
     #[must_use]
-    pub fn colortype(&self) -> ColorType {
+    pub const fn colortype(&self) -> ColorType {
         self.colortype
     }
 
@@ -63,14 +97,24 @@ impl ColorMode {
 
     #[inline(always)]
     #[must_use]
-    pub fn bitdepth(&self) -> u32 {
+    pub const fn bitdepth(&self) -> u32 {
         self.bitdepth
     }
 
     #[inline]
+    #[cfg_attr(debug_assertions, track_caller)]
     pub fn set_bitdepth(&mut self, d: u32) {
-        assert!(d >= 1 && d <= 16);
+        self.try_set_bitdepth(d).unwrap();
+    }
+
+    #[inline]
+    #[cfg_attr(debug_assertions, track_caller)]
+    pub fn try_set_bitdepth(&mut self, d: u32) -> Result<(), Error> {
+        if d < 1 || (d > 8 && d != 16) {
+            return Err(Error::new(37));
+        }
         self.bitdepth = d;
+        Ok(())
     }
 
     /// Set color depth to 8-bit palette and set the colors
@@ -80,7 +124,7 @@ impl ColorMode {
             self.palette_add(c)?;
         }
         self.colortype = ColorType::PALETTE;
-        self.set_bitdepth(8);
+        self.try_set_bitdepth(8)?;
         Ok(())
     }
 
@@ -119,7 +163,13 @@ impl ColorMode {
     #[inline(always)]
     #[must_use]
     pub fn bpp(&self) -> u32 {
-        lodepng_get_bpp_lct(self.colortype, self.bitdepth()) /*4 or 6*/
+        self.colortype.bpp(self.bitdepth)
+    }
+
+    #[inline(always)]
+    #[must_use]
+    pub fn bpp_(&self) -> NonZeroU8 {
+        self.colortype.bpp_(self.bitdepth)
     }
 
     pub(crate) fn clear_key(&mut self) {
@@ -127,7 +177,7 @@ impl ColorMode {
     }
 
     /// `tRNS` chunk
-    #[inline]
+    #[cold]
     pub fn set_key(&mut self, r: u16, g: u16, b: u16) {
         self.key_defined = 1;
         self.key_r = c_uint::from(r);
@@ -193,21 +243,26 @@ impl ColorMode {
     }
 
     /// Returns the byte size of a raw image buffer with given width, height and color mode
-    #[inline]
     #[must_use]
     pub fn raw_size(&self, w: u32, h: u32) -> usize {
-        /*will not overflow for any color type if roughly w * h < 268435455*/
+        self.raw_size_opt(w as _, h as _).expect("overflow")
+    }
+
+    #[inline]
+    fn raw_size_opt(&self, w: u32, h: u32) -> Result<usize, Error> {
         let bpp = self.bpp() as usize;
-        let n = w as usize * h as usize;
-        ((n / 8) * bpp) + ((n & 7) * bpp + 7) / 8
+        let n = (w as usize).checked_mul(h as usize).ok_or(Error::new(77))?;
+        let res = (n / 8).checked_mul(bpp).ok_or(Error::new(77))?.checked_add(((n & 7) * bpp + 7) / 8).ok_or(Error::new(77))?;
+        Ok(res)
     }
 
     /*in an idat chunk, each scanline is a multiple of 8 bits, unlike the lodepng output buffer*/
-    pub(crate) fn raw_size_idat(&self, w: usize, h: usize) -> usize {
-        /*will not overflow for any color type if roughly w * h < 268435455*/
+    pub(crate) fn raw_size_idat(&self, w: u32, h: u32) -> Option<usize> {
+        let w = w as usize;
+        let h = h as usize;
         let bpp = self.bpp() as usize;
-        let line = ((w / 8) * bpp) + ((w & 7) * bpp + 7) / 8;
-        h * line
+        let line = (w / 8).checked_mul(bpp)?.checked_add(((w & 7) * bpp + 7) / 8)?;
+        h.checked_mul(line)
     }
 }
 
@@ -348,7 +403,9 @@ impl Info {
             Ok(c) => c,
             Err(_) => return Err(Error::new(67)),
         };
-        rustimpl::add_chunk(&mut self.unknown_chunks[position as usize], &chtype, data)
+        let mut ch = rustimpl::ChunkBuilder::new(&mut self.unknown_chunks[position as usize], &chtype);
+        ch.extend_from_slice(data)?;
+        ch.finish()
     }
 
     /// Uses linear search to find a given chunk. You can use `b"PLTE"` syntax.
@@ -469,8 +526,9 @@ impl Encoder {
 
     #[inline(always)]
     #[allow(deprecated)]
+    #[track_caller]
     /// Takes any pixel type, but for safety the type has to be marked as "plain old data"
-    pub fn encode<PixelType: rgb::Pod>(&mut self, image: &[PixelType], w: usize, h: usize) -> Result<Vec<u8>, Error> {
+    pub fn encode<PixelType: Pod>(&self, image: &[PixelType], w: usize, h: usize) -> Result<Vec<u8>, Error> {
         if let Some(filters) = &self.predefined_filters {
             if filters.len() < h {
                 return Err(Error::new(88))
@@ -482,7 +540,7 @@ impl Encoder {
     #[inline(always)]
     #[allow(deprecated)]
     /// Takes any pixel type, but for safety the type has to be marked as "plain old data"
-    pub fn encode_file<PixelType: rgb::Pod, P: AsRef<Path>>(&mut self, filepath: P, image: &[PixelType], w: usize, h: usize) -> Result<(), Error> {
+    pub fn encode_file<PixelType: Pod, P: AsRef<Path>>(&self, filepath: P, image: &[PixelType], w: usize, h: usize) -> Result<(), Error> {
         if let Some(filters) = &self.predefined_filters {
             if filters.len() < h {
                 return Err(Error::new(88));
@@ -625,15 +683,17 @@ impl State {
         self.encoder.filter_palette_zero = palette_filter_zero;
     }
 
-    /// See `Encoder'
-    #[deprecated(note = "Use Encoder type instead of State")]
+    /// See `Encoder`
+    #[deprecated(note = "use flate2 crate features instead")]
+    #[allow(deprecated)]
     pub fn set_custom_zlib(&mut self, callback: ffi::custom_compress_callback, context: *const c_void) {
         self.encoder.zlibsettings.custom_zlib = callback;
         self.encoder.zlibsettings.custom_context = context;
     }
 
-    /// See `Encoder'
-    #[deprecated(note = "Use Encoder type instead of State")]
+    /// See `Encoder`
+    #[deprecated(note = "use flate2 crate features instead")]
+    #[allow(deprecated)]
     pub fn set_custom_deflate(&mut self, callback: ffi::custom_compress_callback, context: *const c_void) {
         self.encoder.zlibsettings.custom_deflate = callback;
         self.encoder.zlibsettings.custom_context = context;
@@ -686,17 +746,17 @@ impl State {
             return Err(Error::new(89));
         }
         let iccp = iccp.as_ref().unwrap().data();
-        if iccp.get(0).cloned().unwrap_or(255) == 0 { // text min length is 1
+        if iccp.first().copied().unwrap_or(255) == 0 { // text min length is 1
             return Err(Error::new(89));
         }
 
         let name_len = cmp::min(iccp.len(), 80); // skip name
         for i in 0..name_len {
             if iccp[i] == 0 { // string terminator
-                if iccp.get(i+1).cloned().unwrap_or(255) != 0 { // compression type
+                if iccp.get(i+1).copied().unwrap_or(255) != 0 { // compression type
                     return Err(Error::new(72));
                 }
-                return lodepng_zlib_decompress(&iccp[i+2 ..]);
+                return zlib::decompress(&iccp[i+2 ..], &self.decoder.zlibsettings);
             }
         }
         Err(Error::new(75))
@@ -734,21 +794,26 @@ impl State {
     pub fn inspect(&mut self, input: &[u8]) -> Result<(usize, usize), Error> {
         let (info, w, h) = rustimpl::lodepng_inspect(&self.decoder, input, true)?;
         self.info_png = info;
-        Ok((w, h))
+        Ok((w as _, h as _))
     }
 
     #[deprecated(note = "Use Encoder type instead of State")]
-    pub fn encode<PixelType: rgb::Pod>(&mut self, image: &[PixelType], w: usize, h: usize) -> Result<Vec<u8>, Error> {
+    #[track_caller]
+    pub fn encode<PixelType: Pod>(&self, image: &[PixelType], w: usize, h: usize) -> Result<Vec<u8>, Error> {
+        let w = w.try_into().map_err(|_| Error::new(93))?;
+        let h = h.try_into().map_err(|_| Error::new(93))?;
+
         let image = buffer_for_type(image, w, h, self.info_raw.colortype, self.info_raw.bitdepth)?;
-        rustimpl::lodepng_encode(image, w as c_uint, h as c_uint, self)
+        rustimpl::lodepng_encode(image, w, h, self)
     }
 
     #[deprecated(note = "Use Encoder type instead of State")]
     #[allow(deprecated)]
     #[inline]
-    pub fn encode_file<PixelType: rgb::Pod, P: AsRef<Path>>(&mut self, filepath: P, image: &[PixelType], w: usize, h: usize) -> Result<(), Error> {
+    #[track_caller]
+    pub fn encode_file<PixelType: Pod, P: AsRef<Path>>(&self, filepath: P, image: &[PixelType], w: usize, h: usize) -> Result<(), Error> {
         let buf = self.encode(image, w, h)?;
-        fs::write(filepath, &buf)?;
+        fs::write(filepath, buf)?;
         Ok(())
     }
 }
@@ -766,7 +831,7 @@ impl Default for State {
 }
 
 #[cfg_attr(docsrs, doc(alias = "Gray"))]
-pub use rgb::alt::Gray as Grey;
+pub use rgb::Gray as Grey;
 #[cfg_attr(docsrs, doc(alias = "GrayAlpha"))]
 pub use rgb::alt::GrayAlpha as GreyAlpha;
 
@@ -774,6 +839,8 @@ pub use rgb::alt::GrayAlpha as GreyAlpha;
 ///
 /// Images with >=8bpp are stored with pixel per vec element.
 /// Images with <8bpp are represented as a bunch of bytes, with multiple pixels per byte.
+///
+/// Check `decoder.info_raw()` for more info about the image type.
 #[derive(Debug)]
 pub enum Image {
     /// Bytes of the image. See bpp how many pixels per element there are
@@ -789,11 +856,93 @@ pub enum Image {
     #[cfg_attr(docsrs, doc(alias = "RGBA8"))]
     RGBA(Bitmap<RGBA>),
     #[cfg_attr(docsrs, doc(alias = "RGB8"))]
-    RGB(Bitmap<RGB<u8>>),
-    RGBA16(Bitmap<rgb::RGBA<u16>>),
+    RGB(Bitmap<rgb::Rgb<u8>>),
+    RGBA16(Bitmap<rgb::Rgba<u16>>),
     RGB16(Bitmap<RGB<u16>>),
 }
 
+impl Image {
+    #[must_use] pub fn width(&self) -> usize {
+        match self {
+            Self::RawData(bitmap) => bitmap.width,
+            Self::Grey(bitmap) => bitmap.width,
+            Self::Grey16(bitmap) => bitmap.width,
+            Self::GreyAlpha(bitmap) => bitmap.width,
+            Self::GreyAlpha16(bitmap) => bitmap.width,
+            Self::RGBA(bitmap) => bitmap.width,
+            Self::RGB(bitmap) => bitmap.width,
+            Self::RGBA16(bitmap) => bitmap.width,
+            Self::RGB16(bitmap) => bitmap.width,
+        }
+    }
+
+    #[must_use] pub fn height(&self) -> usize {
+        match self {
+            Self::RawData(bitmap) => bitmap.height,
+            Self::Grey(bitmap) => bitmap.height,
+            Self::Grey16(bitmap) => bitmap.height,
+            Self::GreyAlpha(bitmap) => bitmap.height,
+            Self::GreyAlpha16(bitmap) => bitmap.height,
+            Self::RGBA(bitmap) => bitmap.height,
+            Self::RGB(bitmap) => bitmap.height,
+            Self::RGBA16(bitmap) => bitmap.height,
+            Self::RGB16(bitmap) => bitmap.height,
+        }
+    }
+
+    /// Raw bytes of the underlying buffer
+    #[must_use] pub fn bytes(&self) -> &[u8] {
+        use rgb::bytemuck::cast_slice;
+
+        match self {
+            Self::RawData(bitmap) => {
+                let slice = cast_slice(&bitmap.buffer);
+                slice
+            },
+            Self::Grey(bitmap) => {
+                let slice = cast_slice(&bitmap.buffer);
+                debug_assert_eq!(slice.len(), bitmap.width * bitmap.height);
+                slice
+            },
+            Self::Grey16(bitmap) => {
+                let slice = cast_slice(&bitmap.buffer);
+                debug_assert_eq!(slice.len(), bitmap.width * bitmap.height * 2);
+                slice
+            },
+            Self::GreyAlpha(bitmap) => {
+                let slice = cast_slice(&bitmap.buffer);
+                debug_assert_eq!(slice.len(), bitmap.width * bitmap.height * 2);
+                slice
+            },
+            Self::GreyAlpha16(bitmap) => {
+                let slice = cast_slice(&bitmap.buffer);
+                debug_assert_eq!(slice.len(), bitmap.width * bitmap.height * 4);
+                slice
+            },
+            Self::RGBA(bitmap) => {
+                let slice = cast_slice(&bitmap.buffer);
+                debug_assert_eq!(slice.len(), bitmap.width * bitmap.height * 4);
+                slice
+            },
+            Self::RGB(bitmap) => {
+                let slice = cast_slice(&bitmap.buffer);
+                debug_assert_eq!(slice.len(), bitmap.width * bitmap.height * 3);
+                slice
+            },
+            Self::RGBA16(bitmap) => {
+                let slice = cast_slice(&bitmap.buffer);
+                debug_assert_eq!(slice.len(), bitmap.width * bitmap.height * 8);
+                slice
+            },
+            Self::RGB16(bitmap) => {
+                let slice = cast_slice(&bitmap.buffer);
+                debug_assert_eq!(slice.len(), bitmap.width * bitmap.height * 6);
+                slice
+            },
+        }
+    }
+}
+
 /// Position in the file section after…
 #[derive(Copy, Clone, PartialEq, Eq)]
 pub enum ChunkPosition {
@@ -818,55 +967,59 @@ pub struct Bitmap<PixelType> {
     pub height: usize,
 }
 
-impl<PixelType: rgb::Pod> Bitmap<PixelType> {
+impl<PixelType: Pod> Bitmap<PixelType> {
     /// Convert Vec<u8> to Vec<PixelType>
+    #[cfg_attr(debug_assertions, track_caller)]
     fn from_buffer(buffer: Vec<u8>, width: usize, height: usize) -> Result<Self> {
+        let area = width.checked_mul(height).ok_or(Error::new(77))?;
+
         // Can only cast Vec if alignment doesn't change, and capacity is a round number of pixels
-        let is_safe_to_transmute = unsafe {
-            let whole_capacity_slice = std::slice::from_raw_parts(buffer.as_ptr(), buffer.capacity());
-            let (before, _, after) = whole_capacity_slice.align_to::<PixelType>();
-            1 == std::mem::align_of::<PixelType>() && before.is_empty() && after.is_empty()
-        };
+        let is_safe_to_transmute = 1 == std::mem::align_of::<PixelType>() &&
+            0 == buffer.capacity() % std::mem::align_of::<PixelType>() &&
+            0 == buffer.len() % std::mem::align_of::<PixelType>();
 
         let buffer = if is_safe_to_transmute {
-            debug_assert_eq!(0, buffer.capacity() % std::mem::size_of::<PixelType>());
-            debug_assert!((buffer.len() / std::mem::size_of::<PixelType>()) >= width * height);
             unsafe {
                 let mut buffer = std::mem::ManuallyDrop::new(buffer);
-                Vec::from_raw_parts(buffer.as_mut_ptr() as *mut PixelType,
+                Vec::from_raw_parts(buffer.as_mut_ptr().cast::<PixelType>(),
                     buffer.len() / std::mem::size_of::<PixelType>(),
                     buffer.capacity() / std::mem::size_of::<PixelType>())
             }
         } else {
             // if it's not properly aligned (e.g. reading RGB<u16>), do it the hard way
-            let mut out = Vec::try_with_capacity(width * height)?;
-            assert!(buffer.len() >= width * height * std::mem::size_of::<PixelType>());
+            let mut out = Vec::<PixelType>::new();
+            out.try_reserve_exact(area)?;
+            let dest = out.spare_capacity_mut();
+            assert!(dest.len() >= area);
             unsafe {
-                let mut ptr = buffer.as_ptr().cast::<PixelType>();
-                for _ in 0..width * height {
-                    let px = std::ptr::read_unaligned::<PixelType>(ptr);
-                    ptr = ptr.add(1);
-                    out.push(px);
-                }
+                ptr::copy_nonoverlapping(buffer.as_ptr(), dest.as_mut_ptr().cast::<u8>(), buffer.len());
+                out.set_len(area);
             }
             out
         };
+        if buffer.len() < area {
+            return Err(Error::new(84));
+        }
         Ok(Self { buffer, width, height })
     }
 }
 
 impl<PixelType> fmt::Debug for Bitmap<PixelType> {
+    #[cold]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(f, "{{{} × {} Bitmap}}", self.width, self.height)
     }
 }
 
-fn required_size(w: usize, h: usize, colortype: ColorType, bitdepth: u32) -> usize {
-    colortype.to_color_mode(bitdepth).raw_size(w as c_uint, h as c_uint)
+#[inline]
+fn required_size(w: u32, h: u32, colortype: ColorType, bitdepth: u32) -> Result<usize, Error> {
+    colortype.to_color_mode(bitdepth).raw_size_opt(w, h)
 }
 
-fn new_bitmap(buffer: Vec<u8>, w: usize, h: usize, colortype: ColorType, bitdepth: c_uint) -> Result<Image> {
-    debug_assert!(buffer.len() >= required_size(w, h, colortype, bitdepth));
+fn new_bitmap(buffer: Vec<u8>, w: u32, h: u32, colortype: ColorType, bitdepth: c_uint) -> Result<Image> {
+    debug_assert_eq!(buffer.len(), required_size(w, h, colortype, bitdepth)?);
+    let w = w as usize;
+    let h = h as usize;
 
     Ok(match (colortype, bitdepth) {
         (ColorType::RGBA, 8) => Image::RGBA(Bitmap::from_buffer(buffer, w, h)?),
@@ -877,7 +1030,7 @@ fn new_bitmap(buffer: Vec<u8>, w: usize, h: usize, colortype: ColorType, bitdept
         (ColorType::GREY, 16) => Image::Grey16(Bitmap::from_buffer(buffer, w, h)?),
         (ColorType::GREY_ALPHA, 8) => Image::GreyAlpha(Bitmap::from_buffer(buffer, w, h)?),
         (ColorType::GREY_ALPHA, 16) => Image::GreyAlpha16(Bitmap::from_buffer(buffer, w, h)?),
-        (ColorType::PALETTE, b) | (ColorType::GREY, b) if b > 0 && b <= 8 => Image::RawData(Bitmap {
+        (ColorType::PALETTE | ColorType::GREY, b) if b > 0 && b <= 8 => Image::RawData(Bitmap {
             buffer,
             width: w,
             height: h,
@@ -902,6 +1055,7 @@ fn new_bitmap(buffer: Vec<u8>, w: usize, h: usize, colortype: ColorType, bitdept
 /// * `colortype`: the desired color type for the raw output image. See `ColorType`.
 /// * `bitdepth`: the desired bit depth for the raw output image. 1, 2, 4, 8 or 16. Typically 8.
 #[inline]
+#[cfg_attr(debug_assertions, track_caller)]
 pub fn decode_memory<Bytes: AsRef<[u8]>>(input: Bytes, colortype: ColorType, bitdepth: c_uint) -> Result<Image, Error> {
     let input = input.as_ref();
     assert!(bitdepth > 0 && bitdepth <= 16);
@@ -945,8 +1099,9 @@ pub fn decode24<Bytes: AsRef<[u8]>>(input: Bytes) -> Result<Bitmap<RGB<u8>>, Err
 ///  }
 ///  ```
 #[inline(always)]
+#[cfg_attr(debug_assertions, track_caller)]
 pub fn decode_file<P: AsRef<Path>>(filepath: P, colortype: ColorType, bitdepth: c_uint) -> Result<Image, Error> {
-    decode_memory(&fs::read(filepath)?, colortype, bitdepth)
+    decode_memory(fs::read(filepath)?, colortype, bitdepth)
 }
 
 /// Same as `decode_file`, but always decodes to 32-bit RGBA raw image
@@ -967,22 +1122,26 @@ pub fn decode24_file<P: AsRef<Path>>(filepath: P) -> Result<Bitmap<RGB<u8>>, Err
     }
 }
 
-fn buffer_for_type<PixelType: rgb::Pod>(image: &[PixelType], w: usize, h: usize, colortype: ColorType, bitdepth: u32) -> Result<&[u8], Error> {
-    let bytes_per_pixel = bitdepth as usize/8;
-    assert!(mem::size_of::<PixelType>() <= 4*bytes_per_pixel, "Implausibly large {}-byte pixel data type", mem::size_of::<PixelType>());
+#[cfg_attr(debug_assertions, track_caller)]
+fn buffer_for_type<PixelType: Pod>(image: &[PixelType], w: impl TryInto<u32>, h: impl TryInto<u32>, colortype: ColorType, bitdepth: u32) -> Result<&[u8], Error> {
+    let w = w.try_into().map_err(|_| Error::new(93))?;
+    let h = h.try_into().map_err(|_| Error::new(93))?;
+
+    let required_bytes = required_size(w, h, colortype, bitdepth)?;
 
     let px_bytes = mem::size_of::<PixelType>();
-    let image_bytes = image.len() * px_bytes;
-    let required_bytes = required_size(w, h, colortype, bitdepth);
+    let image_bytes = std::mem::size_of_val(image);
 
     if image_bytes != required_bytes {
-        debug_assert_eq!(image_bytes, required_bytes, "Image is {} bytes large ({}x{}x{}), but needs to be {} ({:?}, {})",
-            image_bytes, w,h,px_bytes, required_bytes, colortype, bitdepth);
+        let bpp = colortype.bpp_(bitdepth).get() as usize;
+        let ch = colortype.channels();
+        assert!(px_bytes == 1 || px_bytes*8 == bpp, "Implausibly large {px_bytes}-byte pixel data type for {bpp}-bit pixels ({ch}ch)");
+        debug_assert_eq!(image_bytes, required_bytes, "Image is {image_bytes} bytes large ({w}x{h} {ch}ch), but needs to be {required_bytes}B ({colortype:?}, {bitdepth})");
         return Err(Error::new(84));
     }
 
     unsafe {
-        Ok(slice::from_raw_parts(image.as_ptr() as *const u8, image_bytes))
+        Ok(slice::from_raw_parts(image.as_ptr().cast::<u8>(), image_bytes))
     }
 }
 
@@ -1000,20 +1159,23 @@ fn buffer_for_type<PixelType: rgb::Pod>(image: &[PixelType], w: usize, h: usize,
 ///
 /// Takes any pixel type, but for safety the type has to be marked as "plain old data"
 #[inline]
-pub fn encode_memory<PixelType: rgb::Pod>(image: &[PixelType], w: usize, h: usize, colortype: ColorType, bitdepth: c_uint) -> Result<Vec<u8>, Error> {
+#[cfg_attr(debug_assertions, track_caller)]
+pub fn encode_memory<PixelType: Pod>(image: &[PixelType], w: usize, h: usize, colortype: ColorType, bitdepth: c_uint) -> Result<Vec<u8>, Error> {
     let image = buffer_for_type(image, w, h, colortype, bitdepth)?;
     rustimpl::lodepng_encode_memory(image, w as u32, h as u32, colortype, bitdepth)
 }
 
 /// Same as `encode_memory`, but always encodes from 32-bit RGBA raw image
 #[inline(always)]
-pub fn encode32<PixelType: rgb::Pod>(image: &[PixelType], w: usize, h: usize) -> Result<Vec<u8>, Error> {
+#[cfg_attr(debug_assertions, track_caller)]
+pub fn encode32<PixelType: Pod>(image: &[PixelType], w: usize, h: usize) -> Result<Vec<u8>, Error> {
     encode_memory(image, w, h, ColorType::RGBA, 8)
 }
 
 /// Same as `encode_memory`, but always encodes from 24-bit RGB raw image
 #[inline(always)]
-pub fn encode24<PixelType: rgb::Pod>(image: &[PixelType], w: usize, h: usize) -> Result<Vec<u8>, Error> {
+#[cfg_attr(debug_assertions, track_caller)]
+pub fn encode24<PixelType: Pod>(image: &[PixelType], w: usize, h: usize) -> Result<Vec<u8>, Error> {
     encode_memory(image, w, h, ColorType::RGB, 8)
 }
 
@@ -1022,21 +1184,24 @@ pub fn encode24<PixelType: rgb::Pod>(image: &[PixelType], w: usize, h: usize) ->
 ///
 /// NOTE: This overwrites existing files without warning!
 #[inline]
-pub fn encode_file<PixelType: rgb::Pod, P: AsRef<Path>>(filepath: P, image: &[PixelType], w: usize, h: usize, colortype: ColorType, bitdepth: c_uint) -> Result<(), Error> {
+#[cfg_attr(debug_assertions, track_caller)]
+pub fn encode_file<PixelType: Pod, P: AsRef<Path>>(filepath: P, image: &[PixelType], w: usize, h: usize, colortype: ColorType, bitdepth: c_uint) -> Result<(), Error> {
     let encoded = encode_memory(image, w, h, colortype, bitdepth)?;
-    fs::write(filepath, &encoded)?;
+    fs::write(filepath, encoded)?;
     Ok(())
 }
 
 /// Same as `encode_file`, but always encodes from 32-bit RGBA raw image
 #[inline(always)]
-pub fn encode32_file<PixelType: rgb::Pod, P: AsRef<Path>>(filepath: P, image: &[PixelType], w: usize, h: usize) -> Result<(), Error> {
+#[cfg_attr(debug_assertions, track_caller)]
+pub fn encode32_file<PixelType: Pod, P: AsRef<Path>>(filepath: P, image: &[PixelType], w: usize, h: usize) -> Result<(), Error> {
     encode_file(filepath, image, w, h, ColorType::RGBA, 8)
 }
 
 /// Same as `encode_file`, but always encodes from 24-bit RGB raw image
 #[inline(always)]
-pub fn encode24_file<PixelType: rgb::Pod, P: AsRef<Path>>(filepath: P, image: &[PixelType], w: usize, h: usize) -> Result<(), Error> {
+#[cfg_attr(debug_assertions, track_caller)]
+pub fn encode24_file<PixelType: Pod, P: AsRef<Path>>(filepath: P, image: &[PixelType], w: usize, h: usize) -> Result<(), Error> {
     encode_file(filepath, image, w, h, ColorType::RGB, 8)
 }
 
@@ -1047,11 +1212,11 @@ pub struct ChunkRef<'a> {
 }
 
 impl<'a> ChunkRef<'a> {
-    #[inline]
+    #[cfg(feature = "c_ffi")]
     pub(crate) unsafe fn from_ptr(data: *const u8) -> Result<Self, Error> {
         debug_assert!(!data.is_null());
         let head = std::slice::from_raw_parts(data, 4);
-        let len = lodepng_chunk_length(head);
+        let len = chunk_length(head);
         let chunk = std::slice::from_raw_parts(data, len + 12);
         Self::new(chunk)
     }
@@ -1060,7 +1225,7 @@ impl<'a> ChunkRef<'a> {
         if data.len() < 12 {
             return Err(Error::new(30));
         }
-        let len = lodepng_chunk_length(data);
+        let len = chunk_length(data);
         /*error: chunk length larger than the max PNG chunk size*/
         if len > (1 << 31) {
             return Err(Error::new(63));
@@ -1078,7 +1243,7 @@ impl<'a> ChunkRef<'a> {
     #[allow(clippy::len_without_is_empty)]
     #[must_use]
     pub fn len(&self) -> usize {
-        lodepng_chunk_length(self.data)
+        chunk_length(self.data)
     }
 
     /// Chunk type, e.g. `tRNS`
@@ -1116,7 +1281,7 @@ impl<'a> ChunkRef<'a> {
 
     #[inline]
     #[must_use]
-    pub fn data(&self) -> &[u8] {
+    pub fn data(&self) -> &'a [u8] {
         let len = self.len();
         &self.data[8..8 + len]
     }
@@ -1133,7 +1298,7 @@ impl<'a> ChunkRef<'a> {
     pub fn check_crc(&self) -> bool {
         let length = self.len();
         /*the CRC is taken of the data and the 4 chunk type letters, not the length*/
-        let crc = rustimpl::lodepng_read32bit_int(&self.data[length + 8..]);
+        let crc = u32::from_be_bytes(self.data[length + 8.. length + 8 + 4].try_into().unwrap());
         let checksum = self.crc();
         crc == checksum
     }
@@ -1152,20 +1317,24 @@ impl<'a> ChunkRef<'a> {
     }
 }
 
+#[deprecated]
+#[allow(dead_code)]
+#[doc(hidden)]
 pub struct ChunkRefMut<'a> {
     data: &'a mut [u8],
 }
 
-impl<'a> ChunkRefMut<'a> {
-    #[inline]
+#[allow(dead_code, deprecated)]
+impl ChunkRefMut<'_> {
+    #[deprecated]
     pub fn data_mut(&mut self) -> &mut [u8] {
         let len = ChunkRef::new(self.data).unwrap().len();
         &mut self.data[8..8 + len]
     }
 
-    #[inline]
+    #[deprecated]
     pub fn generate_crc(&mut self) {
-        rustimpl::lodepng_chunk_generate_crc(self.data)
+        rustimpl::lodepng_chunk_generate_crc(self.data);
     }
 }
 
@@ -1242,6 +1411,11 @@ impl EncoderSettings {
     pub fn new() -> Self {
         Self::default()
     }
+
+    /// Zlib compression level 0=none, 9=best
+    pub fn set_level(&mut self, level: u8) {
+        self.zlibsettings.set_level(level);
+    }
 }
 
 impl Default for EncoderSettings {
@@ -1260,6 +1434,21 @@ impl Default for EncoderSettings {
     }
 }
 
+#[inline]
+fn zero_vec(size: usize) -> Result<Vec<u8>, Error> {
+    let mut vec = Vec::new(); vec.try_reserve_exact(size)?;
+    vec.resize(size, 0);
+    Ok(vec)
+}
+
+#[inline]
+fn try_boxed(bytes: &[u8]) -> Result<Box<[u8]>, Error> {
+    let mut vec = Vec::new();
+    vec.try_reserve_exact(bytes.len())?;
+    vec.extend_from_slice(bytes);
+    Ok(vec.into_boxed_slice())
+}
+
 #[cfg(test)]
 mod test {
     use super::*;
@@ -1271,6 +1460,11 @@ mod test {
         assert_eq!(3, mem::size_of::<RGB<u8>>());
         assert_eq!(2, mem::size_of::<GreyAlpha<u8>>());
         assert_eq!(1, mem::size_of::<Grey<u8>>());
+
+        encode_memory(&[0u8, 1], 4, 4, crate::ColorType::GREY, 1).unwrap();
+        encode_memory(&[0u8, 1], 3, 2, crate::ColorType::GREY, 2).unwrap();
+        encode_memory(&[0u8, 1], 1, 4, crate::ColorType::GREY, 4).unwrap();
+        encode_memory(&[0u8, 1], 2, 1, crate::ColorType::GREY, 8).unwrap();
     }
 
     #[test]
@@ -1310,7 +1504,7 @@ mod test {
             }
 
             let testdata = &[1, 2, 3];
-            info.create_chunk(ChunkPosition::PLTE, &[255, 0, 100, 32], testdata).unwrap();
+            info.create_chunk(ChunkPosition::PLTE, [255, 0, 100, 32], testdata).unwrap();
             assert_eq!(1, info.try_unknown_chunks(ChunkPosition::PLTE).count());
 
             info.create_chunk(ChunkPosition::IHDR, "foob", testdata).unwrap();
